/*
 * fslurp, Copyright (c) 2009 David Stone <ds.fslurp@asif.com>
 *
 * $RCSfile: comm.c,v $
 * $Revision: 1.3 $
 * $Date: 2012/09/02 19:05:13 $
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

#include "fslurp.h"

static char*	SerialPortDevice = DEFAULT_SERIAL_DEVICE;
static char* 	BaudRate = DEFAULT_BAUD_RATE;
static int	InverterFileDesc = -1;

void setupInverterGlobals()
{
	getVersion();
	getActiveInverter();
	getDeviceType();

	InverterGlobals.timestamp = time(NULL);
}

void validateHeader(void)
{
	unsigned char	i;

	for (i = FRONIUS_START_OFFSET; i < FRONIUS_START_LENGTH; i++)
	{
		if (ResponseBuffer[i] != FRONIUS_START_VALUE)
		{
			dieMessage(
				EXIT_COMM_FAILURE,
				"byte %d of the start field invalid: %x\n",
				i, ResponseBuffer[i]);
		}
	}
}

const char* decodeErrorInformation(unsigned char errorInfo)
{
	char*	decodedError;

	switch (errorInfo)
	{
		case FRONIUS_PROTOCOL_ERROR_UNKNOWN_COMMAND:
			decodedError = "unknown command";
			break;
		case FRONIUS_PROTOCOL_ERROR_TIMEOUT:
			decodedError = "timeout";
			break;
		case FRONIUS_PROTOCOL_ERROR_INCORRECT_DATA_STRUCTURE:
			decodedError = "incorrect data structure";
			break;
		case FRONIUS_PROTOCOL_ERROR_COMMAND_QUEUE_FULL:
			decodedError = "command queue full";
			break;
		case FRONIUS_PROTOCOL_ERROR_DEVICE_OR_OPTION_NA:
			decodedError = "device or option not available";
			break;
		case FRONIUS_PROTOCOL_ERROR_NO_RESPONSE:
			decodedError = "no response";
			break;
		case FRONIUS_PROTOCOL_ERROR_SENSOR_ERROR:
			decodedError = "sensor error";
			break;
		case FRONIUS_PROTOCOL_ERROR_SENSOR_NOT_ACTIVE:
			decodedError = "sensor not active";
			break;
		case FRONIUS_PROTOCOL_ERROR_INCORRECT_COMMAND:
			decodedError = "incorrect command";
			break;
		case FRONIUS_PROTOCOL_ERROR_DEVICE_ID_COLLISION:
			decodedError = "device id collision";
			break;
		default:
			// FIX THIS, DS: encode value
			decodedError = "unknown error information";
			break;
	}

	return decodedError;
}

void handleError(void)
{
	struct ErrorResponse*	response;

	response = (struct ErrorResponse*)ResponseBuffer;

	dieMessage(
		EXIT_PROTOCOL_ERROR,
		"command %s failed: %s\n",
		decodeCommand(response->failedCommand),
		decodeErrorInformation(response->errorInformation));
}

void dumpResponseBuffer(void)
{
	unsigned char	length;

	length = ResponseBuffer[FRONIUS_LENGTH_OFFSET];
	length += FRONIUS_NUM_CHECKSUMMED_HDR_BYTES;
	length += FRONIUS_START_LENGTH;
	length += FRONIUS_CHECKSUM_LENGTH;

	hexDump(ResponseBuffer, length);
}

void validateChecksum(void)
{
	unsigned char	length;
	unsigned char	checksum;
	unsigned char	i;

	if (Verbose != 0)
	{
		dumpResponseBuffer();
	}

	length = ResponseBuffer[FRONIUS_LENGTH_OFFSET];
	length += FRONIUS_NUM_CHECKSUMMED_HDR_BYTES;

	checksum = 0;

	for (i = 0; i < length; i++)
	{
		checksum += ResponseBuffer[FRONIUS_LENGTH_OFFSET + i];
	}

	if (checksum != ResponseBuffer[FRONIUS_LENGTH_OFFSET + i])
	{
		dieMessage(
			EXIT_PROTOCOL_ERROR,
			"checksum mismatch, got %d at %d, expected %d\n",
			ResponseBuffer[i],
			i,
			checksum);
	}
}

void readBytes(unsigned char* buffer, int numBytesToRead)
{
	int	totalBytesRead;
	int	bytesRead;
	int	numRetries;

	if (Verbose != 0)
	{
		fprintf(stderr, "readBytes(%d)\n", numBytesToRead);
	}

	for (numRetries = 0, totalBytesRead = 0;
		totalBytesRead < numBytesToRead; )
	{
		bytesRead = read(
			InverterFileDesc,
			buffer + totalBytesRead,
			numBytesToRead - totalBytesRead);

		if (bytesRead == 0)
		{
			numRetries++;
			if (numRetries > numBytesToRead)
			{
				dieMessage(
					EXIT_COMM_FAILURE,
					"readBytes timeout\n");
			}
			else if (Verbose != 0)
			{
				fprintf(stderr, "readBytes: retrying read\n");
			}
		}
		else if (bytesRead == -1)
		{
			dieMessage(
				EXIT_COMM_FAILURE,
				"readBytes failed: %s\n",
				strerror(errno));
		}
		else
		{
			totalBytesRead += bytesRead;
		}
	}
}

const char* decodeCommand(unsigned char command)
{
	char*	decodedCommand;

	switch (command)
	{
		case FRONIUS_CMD_GET_VERSION:
			decodedCommand = "getVersion";
			break;
		case FRONIUS_CMD_GET_DEVICE_TYPE:
			decodedCommand = "getDeviceType";
			break;
		case FRONIUS_CMD_GET_DATE_TIME:
			decodedCommand = "getDateTime";
			break;
		case FRONIUS_CMD_GET_ACTIVE_INVERTER_NUMBER:
			decodedCommand = "getActiveInverterNumber";
			break;
		case FRONIUS_CMD_GET_POWER_NOW:
			decodedCommand = "getPowerNow";
			break;
		case FRONIUS_CMD_GET_ENERGY_TOTAL:
			decodedCommand = "getEnergyTotal";
			break;
		case FRONIUS_CMD_GET_ENERGY_DAY:
			decodedCommand = "getEnergyDay";
			break;
		case FRONIUS_CMD_GET_ENERGY_YEAR:
			decodedCommand = "getEnergyYear";
			break;
		case FRONIUS_CMD_GET_AC_CURRENT_NOW:
			decodedCommand = "getACCurrentNow";
			break;
		case FRONIUS_CMD_GET_AC_VOLTAGE_NOW:
			decodedCommand = "getACVoltageNow";
			break;
		case FRONIUS_CMD_GET_AC_FREQUENCY_NOW:
			decodedCommand = "getACFrequencyNow";
			break;
		case FRONIUS_CMD_GET_DC_CURRENT_NOW:
			decodedCommand = "getDCCurrentNow";
			break;
		case FRONIUS_CMD_GET_DC_VOLTAGE_NOW:
			decodedCommand = "getDCVoltageNow";
			break;
		case FRONIUS_CMD_GET_YIELD_DAY:
			decodedCommand = "getYieldDay";
			break;
		case FRONIUS_CMD_GET_MAXIMUM_POWER_DAY:
			decodedCommand = "getMaximumPowerDay";
			break;
		case FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_DAY:
			decodedCommand = "getMaximumACVoltageDay";
			break;
		case FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_DAY:
			decodedCommand = "getMinimumACVoltageDay";
			break;
		case FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_DAY:
			decodedCommand = "getMaximumDCVoltageDay";
			break;
		case FRONIUS_CMD_GET_OPERATING_HOURS_DAY:
			decodedCommand = "getOperatingHoursDay";
			break;
		case FRONIUS_CMD_GET_YIELD_YEAR:
			decodedCommand = "getYieldYear";
			break;
		case FRONIUS_CMD_GET_MAXIMUM_POWER_TOTAL:
			decodedCommand = "getMaximumPowerTotal";
			break;
		case FRONIUS_CMD_GET_MAXIMUM_AC_VOLTAGE_TOTAL:
			decodedCommand = "getMaximumACVoltageTotal";
			break;
		case FRONIUS_CMD_GET_MINIMUM_AC_VOLTAGE_TOTAL:
			decodedCommand = "getMinimumACVoltageTotal";
			break;
		case FRONIUS_CMD_GET_MAXIMUM_DC_VOLTAGE_TOTAL:
			decodedCommand = "getMaximumDCVoltageTotal";
			break;
		case FRONIUS_CMD_GET_OPERATING_HOURS_TOTAL:
			decodedCommand = "getOperatingHoursTotal";
			break;
		default:
			// FIX THIS, DS: display value
			decodedCommand = "Unknown command";
			break;
	}

	return decodedCommand;
}

void writeBytes(struct Request* request)
{
	int	bytesWritten;
	int	numBytesToWrite;

	assert(request != NULL);
	numBytesToWrite = sizeof(*request);

	if (Verbose != 0)
	{
		fprintf(stderr,
			"writeBytes(%s)\n",
			decodeCommand(request->command));
		hexDump((unsigned char*)request, numBytesToWrite);
	}

	bytesWritten = write(
		InverterFileDesc, (unsigned char*)request, numBytesToWrite);

	if (bytesWritten == -1)
	{
		dieMessage(
			EXIT_COMM_FAILURE,
			"write failed: %s\n",
			strerror(errno));
	}

	if (bytesWritten != numBytesToWrite)
	{
		dieMessage(
			EXIT_COMM_FAILURE,
			"writeBytes(%d) only wrote %d bytes\n",
			numBytesToWrite,
			bytesWritten);
	}
}

void readResponse(void)
{
	unsigned char	length;

	/* read the start sequence and length */

	/* FIX THIS, DS: send error string (header read) as well? */
	readBytes(ResponseBuffer, FRONIUS_HEADER_LENGTH);
	validateHeader();

	length = ResponseBuffer[FRONIUS_LENGTH_OFFSET];

	/* add checksum to length */
	length++;

	/* read the response data and the checksum */
	readBytes(&ResponseBuffer[FRONIUS_HEADER_LENGTH], length);

	validateChecksum();

	if (ResponseBuffer[FRONIUS_COMMAND_OFFSET] == FRONIUS_ERROR)
	{
		handleError();
	}
}

void setSerialDevice(char* device)
{
	SerialPortDevice = device;
}

void setBaudRate(char* baudRate)
{
	if ((strcmp(baudRate, "2400") == 0)
		|| (strcmp(baudRate, "4800") == 0)
		|| (strcmp(baudRate, "9600") == 0)
		|| (strcmp(baudRate, "19200") == 0))
	{
		BaudRate = baudRate;
	}
	else
	{
		dieUsage("%s is not a supported baud rate\n.", baudRate);
	}
}

unsigned int getBaudRateFlag()
{
	unsigned int	baudRateFlag;

	if (strcmp(BaudRate, "2400") == 0)
	{
		baudRateFlag = B2400;
	}
	else if (strcmp(BaudRate, "4800") == 0)
	{
		baudRateFlag = B4800;
	}
	else if (strcmp(BaudRate, "9600") == 0)
	{
		baudRateFlag = B9600;
	}
	else
	{
		assert(strcmp(BaudRate, "19200") == 0);
		baudRateFlag = B19200;
	}

	return baudRateFlag;
}

void setupSerialPort(void)
{
	struct termios	termio;

	if ((InverterFileDesc = open(SerialPortDevice, O_RDWR|O_NOCTTY)) == -1)
	{
		dieMessage(
			EXIT_COMM_FAILURE,
			"open of %s failed: %s\n",
			SerialPortDevice,
			strerror(errno));
	}

	if (tcflush(InverterFileDesc, TCIFLUSH) == -1)
	{
		dieMessage(
			EXIT_COMM_FAILURE,
			"tcflush failed: %s\n",
			strerror(errno));
	}

	memset(&termio, 0, sizeof (termio));

	termio.c_cflag = getBaudRateFlag() | CS8 | CLOCAL | CREAD;
	termio.c_iflag = IGNPAR;
	termio.c_oflag = 0;
	termio.c_lflag = 0;
	termio.c_cc[VMIN] = 0;
	termio.c_cc[VTIME] = 1;

	if (tcsetattr(InverterFileDesc, TCSANOW, &termio) == -1)
	{
		dieMessage(
			EXIT_COMM_FAILURE,
			"tcsetattr failed: %s\n",
			strerror(errno));
	}
}

void flushSerialPort(void)
{
	if (tcflush(InverterFileDesc, TCIOFLUSH) == -1)
	{
		dieMessage(
			EXIT_COMM_FAILURE,
			"tcflush(TCIOFLUSH) failed: %s\n",
			strerror(errno));
	}
}
