/*
 * fslurp, Copyright (c) 2009 David Stone <ds.fslurp@asif.com>
 *
 * $RCSfile: main.c,v $
 * $Revision: 1.4 $
 * $Date: 2010/05/26 17:54:12 $
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

#include "fslurp.h"

static char*	ProgName;

void usage(void)
{
	fprintf(stderr,
		"Usage: %s [-b baud][-d delimiter][-H][-h]", ProgName);
	fprintf(stderr, "[-p device]\n");
	fprintf(stderr, "\t\t[-r report][-t format][-v]\n");
	fprintf(stderr, "\n");

	fprintf(stderr,
		"\t-b baud	Set the baud rate.  Supported rates are:\n");
	fprintf(stderr,
		"\t		2400, 4800, 9600, and 19200. Defaults to %s.\n",
				DEFAULT_BAUD_RATE);

	fprintf(stderr,
		"\t-d delimter	Set delimiter for CSV-style output. The\n");
	fprintf(stderr,
		"\t		default is no delimiter, human-readable.\n");

	fprintf(stderr,
		"\t-H		Display the delimited output header for the\n");
	fprintf(stderr,
		"\t		current reporting mode.  The -d option must\n");
	fprintf(stderr,
		"\t		also be set.\n");

	fprintf(stderr,
		"\t-h		Display this help message.\n");

	fprintf(stderr,
		"\t-p device	Set the serial port device. Defaults to %s.\n",
					DEFAULT_SERIAL_DEVICE);

	fprintf(stderr,
		"\t-r report	Set report type. Supported types are:\n");
	fprintf(stderr,
		"\t		now, day, total, and all  Defaults to %s.\n",
				DEFAULT_REPORT_TYPE);

	fprintf(stderr,
		"\t-t format	Set the format for the timestamp field,\n");
	fprintf(stderr,
		"\t		using the Linux strftime(3) format string.\n");
	fprintf(stderr,
		"\t		The default is %s which looks like:\n", "%c");
	fprintf(stderr,
		"\t\t		Wed May 26 10:43:25 2010\n");

	fprintf(stderr,
		"\t-v		Enable verbose, debugging output.\n");

	fprintf(stderr, "\n");
	fprintf(stderr, "Return codes:\n");
	fprintf(stderr, "	%d	Success.\n",
		EXIT_NORMAL);
	fprintf(stderr, "	%d	Communications failure.\n",
		EXIT_COMM_FAILURE);
	fprintf(stderr, "	%d	No active inverter found.\n",
		EXIT_NO_INVERTER);
	fprintf(stderr, "	%d	Usage error.\n",
		EXIT_USAGE_ERROR);
	fprintf(stderr, "	%d	Fronius protocol error.\n",
		EXIT_PROTOCOL_ERROR);
}

void dieUsage(char* format, ...)
{
	va_list	ap;
	va_start(ap, format);
	fprintf(stderr, "%s ERROR: ", ProgName);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, "\n");
	usage();
	exit(EXIT_USAGE_ERROR);
}

void dieMessage(int returnCode, char* format, ...)
{
	va_list	ap;
	va_start(ap, format);
	fprintf(stderr, "%s ERROR: ", ProgName);
	vfprintf(stderr, format, ap);
	va_end(ap);
	exit(returnCode);
}

void inverterReport(void)
{
	struct InverterNowData		nowData;
	struct InverterDayData		dayData;
	struct InverterTotalData	totalData;

	switch (ReportType)
	{
		case ReportTypeNow:
			getInverterNowData(&nowData);
			inverterNowReport(nowData);
			break;
		case ReportTypeDay:
			getInverterDayData(&dayData);
			inverterDayReport(dayData);
			break;
		case ReportTypeTotal:
			getInverterTotalData(&totalData);
			inverterTotalReport(totalData);
			break;
		case ReportTypeAll:
			getInverterNowData(&nowData);
			getInverterDayData(&dayData);
			getInverterTotalData(&totalData);
			inverterAllReport(nowData, dayData, totalData);
			break;
		default:
			assert(0);
	}
}

void setReportType(const char* reportType)
{
	if (strcmp(reportType, "now") == 0)
	{
		ReportType = ReportTypeNow;
	}
	else if (strcmp(reportType, "day") == 0)
	{
		ReportType = ReportTypeDay;
	}
	else if (strcmp(reportType, "total") == 0)
	{
		ReportType = ReportTypeTotal;
	}
	else if (strcmp(reportType, "all") == 0)
	{
		ReportType = ReportTypeAll;
	}
	else
	{
		dieUsage("%s is not a supported report type.\n", reportType);
	}
}

void doArgs(int argc, char* argv[])
{
	int	c;

	assert(argc >= 1);
	ProgName = argv[0];

	while (1)
	{
		c = getopt(argc, argv, "?b:d:Hhp:r:t:v");

		if (c == -1)
		{
			break;
		}

		switch (c)
		{
			case 'b':
				setBaudRate(optarg);
				break;
			case 'd':
				Delimiter = optarg;
				break;
			case 'H':
				HeaderReport = 1;
				break;
			case 'h':
			case '?':
				usage();
				exit(EXIT_NORMAL);
				break;
			case 'p':
				setSerialDevice(optarg);
				break;
			case 'r':
				setReportType(optarg);
				break;
			case 't':
				TimeFormat = optarg;
				break;
			case 'v':
				Verbose = 1;
				break;
			default:
				dieUsage("Unsupported option: %c\n", c);
				break;
		}
	}

	if (optind < argc)
	{
		dieUsage("Unsupported arguments.\n");
	}

	if ((HeaderReport == 1) && (Delimiter == NULL))
	{
		dieUsage("-H requires that the delimiter also be set (-d).\n");
	}
}

int main(int argc, char* argv[])
{
	doArgs(argc, argv);

	setupSerialPort();
	setupInverterGlobals();

	if (HeaderReport == 1)
	{
		inverterHeaderReport();
	}
	else
	{
		inverterReport();
	}

	return 0;
}
